@@ -13,10 +13,30 @@
 """SSH agent client"""
 
 import asyncio
+import errno
 import os
+import sys
+import tempfile
 
 import asyncssh
 
+from .logging import logger
+
+
+try:
+    if sys.platform == 'win32': # pragma: no cover
+        from .agent_win32 import open_agent
+    else:
+        from .agent_unix import open_agent
+except ImportError as exc: # pragma: no cover
+    def open_agent(loop, agent_path, reason=str(exc)):
+        """Dummy function if we're unable to import agent support"""
+
+        # pylint: disable=unused-argument
+
+        raise OSError(errno.ENOENT, 'Agent support unavailable: %s' % reason)
+
+from .listener import create_unix_forward_listener
 from .misc import ChannelOpenError, load_default_keypairs
 from .packet import Byte, String, UInt32, PacketDecodeError, SSHPacket
 from .public_key import SSHKeyPair
@@ -57,6 +77,26 @@ SSH_AGENT_RSA_SHA2_512                   = 4
 # pylint: enable=bad-whitespace
 
 
+class _X11AgentListener:
+    """Listener used to forward agent connections"""
+
+    def __init__(self, tempdir, path, unix_listener):
+        self._tempdir = tempdir
+        self._path = path
+        self._unix_listener = unix_listener
+
+    def get_path(self):
+        """Return the path being listened on"""
+
+        return self._path
+
+    def close(self):
+        """Close the agent listener"""
+
+        self._unix_listener.close()
+        self._tempdir.cleanup()
+
+
 class SSHAgentKeyPair(SSHKeyPair):
     """Surrogate for a key managed by the SSH agent"""
 
@@ -122,7 +162,7 @@ class SSHAgentClient:
         self._agent_path = agent_path
         self._reader = None
         self._writer = None
-        self._lock = asyncio.Lock()
+        self._lock = asyncio.Lock(loop=loop)
 
     def _cleanup(self):
         """Clean up this SSH agent client"""
@@ -150,15 +190,12 @@ class SSHAgentClient:
     def connect(self):
         """Connect to the SSH agent"""
 
-        if isinstance(self._agent_path, str):
-            # pylint doesn't think open_unix_connection exists
-            # pylint: disable=no-member
-            self._reader, self._writer = \
-                yield from asyncio.open_unix_connection(self._agent_path,
-                                                        loop=self._loop)
-        else:
+        if isinstance(self._agent_path, asyncssh.SSHServerConnection):
             self._reader, self._writer = \
                 yield from self._agent_path.open_agent_connection()
+        else:
+            self._reader, self._writer = \
+                yield from open_agent(self._loop, self._agent_path)
 
     @asyncio.coroutine
     def _make_request(self, msgtype, *args):
@@ -510,19 +547,26 @@ def connect_agent(agent_path=None, *, loop=None):
 
     """
 
-    if not loop:
-        loop = asyncio.get_event_loop()
-
-    if not agent_path:
-        agent_path = os.environ.get('SSH_AUTH_SOCK', None)
-
-        if not agent_path:
-            return None
-
     agent = SSHAgentClient(loop, agent_path)
 
     try:
         yield from agent.connect()
         return agent
-    except (OSError, ChannelOpenError):
+    except (OSError, ChannelOpenError) as exc:
+        logger.debug('Unable to contact agent: %s', exc)
+        return None
+
+
+@asyncio.coroutine
+def create_agent_listener(conn, loop):
+    """Create a listener for forwarding ssh-agent connections"""
+
+    try:
+        tempdir = tempfile.TemporaryDirectory(prefix='asyncssh-')
+        path = os.path.join(tempdir.name, 'agent')
+        unix_listener = yield from create_unix_forward_listener(
+            conn, loop, conn.create_agent_connection, path)
+
+        return _X11AgentListener(tempdir, path, unix_listener)
+    except OSError:
         return None
