@@ -1,4 +1,4 @@
-# Copyright (c) 2014-2015 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2014-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -16,10 +16,11 @@ from .cipher import register_cipher, lookup_cipher
 
 from .ec import lookup_ec_curve_by_params
 
-# Import PyCA versions of DSA, ECDSA, and RSA
+# Import PyCA versions of DSA, ECDSA, RSA, and PBKDF2
 from .pyca.dsa import DSAPrivateKey, DSAPublicKey
 from .pyca.ec import ECDSAPrivateKey, ECDSAPublicKey, ECDH
 from .pyca.rsa import RSAPrivateKey, RSAPublicKey
+from .pyca.kdf import pbkdf2_hmac
 
 # Import pyca module to get ciphers defined there registered
 from . import pyca
@@ -38,3 +39,10 @@ try:
     from .umac import umac32, umac64, umac96, umac128
 except (ImportError, AttributeError, OSError): # pragma: no cover
     pass
+
+# Import X.509 certificate support if available
+try:
+    from .pyca.x509 import X509Name, X509NamePattern
+    from .pyca.x509 import generate_x509_certificate, import_x509_certificate
+except ImportError: # pragma: no cover
+    pass
