@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2015 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2013-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -15,10 +15,10 @@
 from .asn1 import ASN1DecodeError, ObjectIdentifier, der_encode, der_decode
 from .crypto import DSAPrivateKey, DSAPublicKey
 from .misc import all_ints
-from .packet import MPInt, String, PacketDecodeError, SSHPacket
-from .public_key import SSHKey, SSHCertificateV01
-from .public_key import KeyExportError
+from .packet import MPInt
+from .public_key import SSHKey, SSHOpenSSHCertificateV01, KeyExportError
 from .public_key import register_public_key_alg, register_certificate_alg
+from .public_key import register_x509_certificate_alg
 
 # Short variable names are used here, matching names in the spec
 # pylint: disable=invalid-name
@@ -30,9 +30,9 @@ class _DSAKey(SSHKey):
     algorithm = b'ssh-dss'
     pem_name = b'DSA'
     pkcs8_oid = ObjectIdentifier('1.2.840.10040.4.1')
-
-    def __init__(self, key):
-        self._key = key
+    sig_algorithms = (algorithm,)
+    x509_algorithms = (b'x509v3-' + algorithm,)
+    all_sig_algorithms = set(sig_algorithms)
 
     def __eq__(self, other):
         # This isn't protected access - both objects are _DSAKey instances
@@ -49,17 +49,25 @@ class _DSAKey(SSHKey):
         return hash((self._key.p, self._key.q, self._key.g,
                      self._key.y, self._key.x))
 
+    @classmethod
+    def generate(cls, algorithm):
+        """Generate a new DSA private key"""
+
+        # pylint: disable=unused-argument
+
+        return cls(DSAPrivateKey.generate(key_size=1024))
+
     @classmethod
     def make_private(cls, *args):
         """Construct a DSA private key"""
 
-        return cls(DSAPrivateKey(*args))
+        return cls(DSAPrivateKey.construct(*args))
 
     @classmethod
     def make_public(cls, *args):
         """Construct a DSA public key"""
 
-        return cls(DSAPublicKey(*args))
+        return cls(DSAPublicKey.construct(*args))
 
     @classmethod
     def decode_pkcs1_private(cls, key_data):
@@ -181,41 +189,53 @@ class _DSAKey(SSHKey):
         return b''.join((MPInt(self._key.p), MPInt(self._key.q),
                          MPInt(self._key.g), MPInt(self._key.y)))
 
-    def sign(self, data):
-        """Return a signature of the specified data using this key"""
+    def encode_agent_cert_private(self):
+        """Encode DSA certificate private key data for agent"""
+
+        if not self._key.x:
+            raise KeyExportError('Key is not private')
+
+        return MPInt(self._key.x)
+
+    def sign_der(self, data, sig_algorithm):
+        """Compute a DER-encoded signature of the specified data"""
+
+        # pylint: disable=unused-argument
 
         if not self._key.x:
             raise ValueError('Private key needed for signing')
 
-        r, s = self._key.sign(data)
-        return b''.join((String(self.algorithm),
-                         String(r.to_bytes(20, 'big') +
-                                s.to_bytes(20, 'big'))))
+        return self._key.sign(data)
 
-    def verify(self, data, sig):
-        """Verify a signature of the specified data using this key"""
+    def verify_der(self, data, sig_algorithm, sig):
+        """Verify a DER-encoded signature of the specified data"""
 
-        try:
-            packet = SSHPacket(sig)
+        # pylint: disable=unused-argument
 
-            if packet.get_string() != self.algorithm:
-                return False
+        return self._key.verify(data, sig)
 
-            sig = packet.get_string()
-            packet.check_end()
+    def sign_ssh(self, data, sig_algorithm):
+        """Compute an SSH-encoded signature of the specified data"""
 
-            if len(sig) != 40:
-                return False
+        r, s = der_decode(self.sign_der(data, sig_algorithm))
+        return r.to_bytes(20, 'big') + s.to_bytes(20, 'big')
 
-            r = int.from_bytes(sig[:20], 'big')
-            s = int.from_bytes(sig[20:], 'big')
+    def verify_ssh(self, data, sig_algorithm, sig):
+        """Verify an SSH-encoded signature of the specified data"""
 
-            return self._key.verify(data, (r, s))
-        except PacketDecodeError:
+        if len(sig) != 40:
             return False
 
+        r = int.from_bytes(sig[:20], 'big')
+        s = int.from_bytes(sig[20:], 'big')
+
+        return self.verify_der(data, sig_algorithm, der_encode((r, s)))
+
 
 register_public_key_alg(b'ssh-dss', _DSAKey)
 
-register_certificate_alg(b'ssh-dss-cert-v01@openssh.com',
-                         _DSAKey, SSHCertificateV01)
+register_certificate_alg(1, b'ssh-dss', b'ssh-dss-cert-v01@openssh.com',
+                         _DSAKey, SSHOpenSSHCertificateV01)
+
+for alg in _DSAKey.x509_algorithms:
+    register_x509_certificate_alg(alg)
