@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2016 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2013-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -41,9 +41,9 @@ randrange = _random.randrange
 
 # Avoid deprecation warning for asyncio.async()
 if python344:
-    ensure_future = asyncio.ensure_future
+    create_task = asyncio.ensure_future
 else: # pragma: no cover
-    ensure_future = asyncio.async    # pylint: disable=deprecated-method
+    create_task = asyncio.async    # pylint: disable=no-member
 
 
 def all_ints(seq):
@@ -64,6 +64,10 @@ def load_default_keypairs(passphrase=None):
         try:
             file = os.path.join(os.path.expanduser('~'), '.ssh', file)
             result.extend(asyncssh.load_keypairs(file, passphrase))
+        except asyncssh.KeyImportError as exc:
+            # Ignore encrypted default keys if a passphrase isn't provided
+            if not str(exc).startswith('Passphrase'):
+                raise
         except OSError:
             pass
 
