@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2015 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2013-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -19,7 +19,8 @@ from .misc import BreakReceived, SignalReceived, TerminalSizeChanged
 from .misc import async_iterator, python35
 from .session import SSHClientSession, SSHServerSession
 from .session import SSHTCPSession, SSHUNIXSession
-from .sftp import SFTPServerHandler
+from .sftp import run_sftp_server
+from .scp import run_scp_server
 
 _NEWLINE = object()
 
@@ -280,47 +281,47 @@ class SSHStreamSession:
         self._connection_lost = False
         self._recv_buf = {None: []}
         self._recv_buf_len = 0
-        self._read_waiter = {None: None}
+        self._read_waiters = {None: None}
         self._read_paused = False
         self._write_paused = False
-        self._drain_waiters = []
+        self._drain_waiters = {None: set()}
 
     @asyncio.coroutine
     def _block_read(self, datatype):
         """Wait for more data to arrive on the stream"""
 
-        if self._read_waiter[datatype]:
+        if self._read_waiters[datatype]:
             raise RuntimeError('read called while another coroutine is '
                                'already waiting to read')
 
         try:
             waiter = asyncio.Future(loop=self._loop)
-            self._read_waiter[datatype] = waiter
+            self._read_waiters[datatype] = waiter
             yield from waiter
         finally:
-            self._read_waiter[datatype] = None
+            self._read_waiters[datatype] = None
 
     def _unblock_read(self, datatype):
         """Signal that more data has arrived on the stream"""
 
-        waiter = self._read_waiter[datatype]
-        if waiter:
-            if not waiter.done():
-                waiter.set_result(None)
+        waiter = self._read_waiters[datatype]
+        if waiter and not waiter.done():
+            waiter.set_result(None)
 
     def _should_block_drain(self, datatype):
         """Return whether output is still being written to the channel"""
 
         # pylint: disable=unused-argument
 
-        return self._write_paused
+        return self._write_paused and not self._connection_lost
 
-    def _unblock_drain(self):
+    def _unblock_drain(self, datatype):
         """Signal that more data can be written on the stream"""
 
-        for waiter in self._drain_waiters:
-            if not waiter.done(): # pragma: no branch
-                waiter.set_result(None)
+        if not self._should_block_drain(datatype):
+            for waiter in self._drain_waiters[datatype]:
+                if not waiter.done(): # pragma: no branch
+                    waiter.set_result(None)
 
     def _should_pause_reading(self):
         """Return whether to pause reading from the channel"""
@@ -358,7 +359,10 @@ class SSHStreamSession:
 
         for datatype in chan.get_read_datatypes():
             self._recv_buf[datatype] = []
-            self._read_waiter[datatype] = None
+            self._read_waiters[datatype] = None
+
+        for datatype in chan.get_write_datatypes():
+            self._drain_waiters[datatype] = set()
 
     def connection_lost(self, exc):
         """Handle an incoming channel close"""
@@ -368,13 +372,13 @@ class SSHStreamSession:
 
         if not self._eof_received:
             if exc:
-                for datatype in self._read_waiter.keys():
+                for datatype in self._read_waiters:
                     self._recv_buf[datatype].append(exc)
 
             self.eof_received()
 
-        if self._write_paused:
-            self._unblock_drain()
+        for datatype in self._drain_waiters:
+            self._unblock_drain(datatype)
 
     def data_received(self, data, datatype):
         """Handle incoming data on the channel"""
@@ -389,7 +393,7 @@ class SSHStreamSession:
 
         self._eof_received = True
 
-        for datatype in self._read_waiter.keys():
+        for datatype in self._read_waiters:
             self._unblock_read(datatype)
 
         return True
@@ -408,7 +412,9 @@ class SSHStreamSession:
         """Handle a request to resume writing on the channel"""
 
         self._write_paused = False
-        self._unblock_drain()
+
+        for datatype in self._drain_waiters:
+            self._unblock_drain(datatype)
 
     @asyncio.coroutine
     def read(self, n, datatype, exact):
@@ -506,13 +512,13 @@ class SSHStreamSession:
     def drain(self, datatype):
         """Wait for data written to the channel to drain"""
 
-        while self._should_block_drain(datatype) and not self._connection_lost:
+        while self._should_block_drain(datatype):
             try:
                 waiter = asyncio.Future(loop=self._loop)
-                self._drain_waiters.append(waiter)
+                self._drain_waiters[datatype].add(waiter)
                 yield from waiter
             finally:
-                self._drain_waiters.remove(waiter)
+                self._drain_waiters[datatype].remove(waiter)
 
         if self._connection_lost:
             exc = self._exception
@@ -531,11 +537,12 @@ class SSHClientStreamSession(SSHStreamSession, SSHClientSession):
 class SSHServerStreamSession(SSHStreamSession, SSHServerSession):
     """SSH server stream session handler"""
 
-    def __init__(self, session_factory, sftp_factory):
+    def __init__(self, session_factory, sftp_factory, allow_scp):
         super().__init__()
 
         self._session_factory = session_factory
         self._sftp_factory = sftp_factory
+        self._allow_scp = allow_scp and bool(sftp_factory)
 
     def shell_requested(self):
         """Return whether a shell can be requested"""
@@ -545,7 +552,8 @@ class SSHServerStreamSession(SSHStreamSession, SSHServerSession):
     def exec_requested(self, command):
         """Return whether execution of a command can be requested"""
 
-        return bool(self._session_factory)
+        return ((self._allow_scp and command.startswith('scp ')) or
+                bool(self._session_factory))
 
     def subsystem_requested(self, subsystem):
         """Return whether starting a subsystem can be requested"""
@@ -558,18 +566,26 @@ class SSHServerStreamSession(SSHStreamSession, SSHServerSession):
     def session_started(self):
         """Start a session for this newly opened server channel"""
 
+        command = self._chan.get_command()
+
+        stdin = SSHReader(self, self._chan)
+        stdout = SSHWriter(self, self._chan)
+        stderr = SSHWriter(self, self._chan, EXTENDED_DATA_STDERR)
+
         if self._chan.get_subsystem() == 'sftp':
             self._chan.set_encoding(None)
             self._encoding = None
 
-            handler = SFTPServerHandler(self._sftp_factory, self._conn,
-                                        SSHReader(self, self._chan),
-                                        SSHWriter(self, self._chan)).start()
+            handler = run_sftp_server(self._sftp_factory(self._conn),
+                                      stdin, stdout)
+        elif self._allow_scp and command and command.startswith('scp '):
+            self._chan.set_encoding(None)
+            self._encoding = None
+
+            handler = run_scp_server(self._sftp_factory(self._conn),
+                                     command, stdin, stdout, stderr)
         else:
-            handler = self._session_factory(SSHReader(self, self._chan),
-                                            SSHWriter(self, self._chan),
-                                            SSHWriter(self, self._chan,
-                                                      EXTENDED_DATA_STDERR))
+            handler = self._session_factory(stdin, stdout, stderr)
 
         if asyncio.iscoroutine(handler):
             self._conn.create_task(handler)
