@@ -1,6 +1,6 @@
-#!/usr/bin/env python3.4
+#!/usr/bin/env python3.5
 #
-# Copyright (c) 2013-2015 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2013-2016 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -23,17 +23,16 @@ class MySSHTCPSession(asyncssh.SSHTCPSession):
         if exc:
             print('Direct connection error:', str(exc), file=sys.stderr)
 
-@asyncio.coroutine
-def run_client():
-    with (yield from asyncssh.connect('localhost')) as conn:
-        chan, session = yield from conn.create_connection(MySSHTCPSession,
-                                                          'www.google.com', 80)
+async def run_client():
+    async with asyncssh.connect('localhost') as conn:
+        chan, session = await conn.create_connection(MySSHTCPSession,
+                                                     'www.google.com', 80)
 
         # By default, TCP connections send and receive bytes
         chan.write(b'HEAD / HTTP/1.0\r\n\r\n')
         chan.write_eof()
 
-        yield from chan.wait_closed()
+        await chan.wait_closed()
 
 try:
     asyncio.get_event_loop().run_until_complete(run_client())
