@@ -1,6 +1,6 @@
-#!/usr/bin/env python3.4
+#!/usr/bin/env python3.5
 #
-# Copyright (c) 2015 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2015-2016 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -14,11 +14,10 @@
 
 import asyncio, asyncssh, sys
 
-@asyncio.coroutine
-def run_client():
-    with (yield from asyncssh.connect('localhost')) as conn:
-        with (yield from conn.start_sftp_client()) as sftp:
-            yield from sftp.get('example.txt')
+async def run_client():
+    async with asyncssh.connect('localhost') as conn:
+        async with conn.start_sftp_client() as sftp:
+            await sftp.get('example.txt')
 
 try:
     asyncio.get_event_loop().run_until_complete(run_client())
