@@ -1,4 +1,4 @@
-# Copyright (c) 2015-2016 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2015-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -13,37 +13,27 @@
 """Unit tests for authentication"""
 
 import asyncio
+import unittest
 
 import asyncssh
 
 from asyncssh.auth import MSG_USERAUTH_PK_OK, lookup_client_auth
 from asyncssh.auth import get_server_auth_methods, lookup_server_auth
+from asyncssh.auth import MSG_USERAUTH_GSSAPI_RESPONSE
 from asyncssh.constants import MSG_USERAUTH_REQUEST, MSG_USERAUTH_FAILURE
 from asyncssh.constants import MSG_USERAUTH_SUCCESS
+from asyncssh.gss import GSSClient, GSSServer
 from asyncssh.misc import DisconnectError, PasswordChangeRequired
 from asyncssh.packet import SSHPacket, Boolean, Byte, NameList, String
 from asyncssh.public_key import SSHLocalKeyPair
 
-from .util import asynctest, create_task, AsyncTestCase, ConnectionStub
+from .util import asynctest, gss_available, patch_gss
+from .util import AsyncTestCase, ConnectionStub
 
 
 class _AuthConnectionStub(ConnectionStub):
     """Connection stub class to test authentication"""
 
-    @asyncio.coroutine
-    def _run_task(self, coro):
-        """Run an asynchronous task"""
-
-        try:
-            yield from coro
-        except DisconnectError as exc:
-            self.connection_lost(exc)
-
-    def create_task(self, coro):
-        """Create an asynchronous task"""
-
-        return create_task(self._run_task(coro))
-
     def connection_lost(self, exc):
         """Handle the closing of a connection"""
 
@@ -54,6 +44,20 @@ class _AuthConnectionStub(ConnectionStub):
 
         raise NotImplementedError
 
+    def _get_userauth_request_packet(self, method, args):
+        """Get packet data for a user authentication request"""
+
+        # pylint: disable=no-self-use
+
+        return b''.join((Byte(MSG_USERAUTH_REQUEST), String('user'),
+                         String('service'), String(method)) + args)
+
+    def get_userauth_request_data(self, method, *args):
+        """Get signature data for a user authentication request"""
+
+        return String('') + self._get_userauth_request_packet(method, args)
+
+
 class _AuthClientStub(_AuthConnectionStub):
     """Stub class for client connection"""
 
@@ -64,17 +68,20 @@ class _AuthClientStub(_AuthConnectionStub):
         client_conn = cls(method, **kwargs)
         return client_conn, client_conn.get_peer()
 
-    def __init__(self, method, public_key_auth=False, client_key=None,
-                 client_cert=None, override_pk_ok=False, password_auth=False,
-                 password=None, password_change=NotImplemented,
-                 password_change_prompt=None, kbdint_auth=False,
-                 kbdint_submethods=None, kbdint_challenge=None,
-                 kbdint_response=None, success=False):
-        super().__init__(_AuthServerStub(self, public_key_auth, override_pk_ok,
+    def __init__(self, method, gss_host=None, override_gss_mech=False,
+                 public_key_auth=False, client_key=None, client_cert=None,
+                 override_pk_ok=False, password_auth=False, password=None,
+                 password_change=NotImplemented, password_change_prompt=None,
+                 kbdint_auth=False, kbdint_submethods=None,
+                 kbdint_challenge=None, kbdint_response=None, success=False):
+        super().__init__(_AuthServerStub(self, gss_host, override_gss_mech,
+                                         public_key_auth, override_pk_ok,
                                          password_auth, password_change_prompt,
                                          kbdint_auth, kbdint_challenge,
                                          success), False)
 
+        self._gss = GSSClient(gss_host, False) if gss_host else None
+
         self._client_key = client_key
         self._client_cert = client_cert
 
@@ -90,13 +97,15 @@ class _AuthClientStub(_AuthConnectionStub):
 
         if self._auth is None:
             self.close()
-            self._peer.close()
             raise ValueError('Invalid auth method')
 
-    def connection_lost(self, exc):
+    def connection_lost(self, exc=None):
         """Handle the closing of a connection"""
 
-        self._auth_waiter.set_exception(exc)
+        if exc:
+            self._auth_waiter.set_exception(exc)
+
+        self.close()
 
     def process_packet(self, data):
         """Process an incoming packet"""
@@ -126,10 +135,7 @@ class _AuthClientStub(_AuthConnectionStub):
             self._auth = None
             self._auth_waiter = None
         else:
-            try:
-                self._auth.process_packet(pkttype, packet)
-            except DisconnectError as exc:
-                self.connection_lost(exc)
+            self._auth.process_packet(pkttype, packet)
 
     def get_auth_result(self):
         """Return the result of the authentication"""
@@ -148,14 +154,23 @@ class _AuthClientStub(_AuthConnectionStub):
     def send_userauth_request(self, method, *args, key=None):
         """Send a user authentication request"""
 
-        packet = b''.join((Byte(MSG_USERAUTH_REQUEST), String('user'),
-                           String(b'service'), String(method)) + args)
+        packet = self._get_userauth_request_packet(method, args)
 
         if key:
             packet += String(key.sign(String('') + packet))
 
         self.send_packet(packet)
 
+    def get_gss_context(self):
+        """Return the GSS context associated with this connection"""
+
+        return self._gss
+
+    def gss_mic_auth_requested(self):
+        """Return whether to allow GSS MIC authentication or not"""
+
+        return bool(self._gss)
+
     @asyncio.coroutine
     def public_key_auth_requested(self):
         """Return key to use for public key authentication"""
@@ -215,11 +230,15 @@ class _AuthClientStub(_AuthConnectionStub):
 class _AuthServerStub(_AuthConnectionStub):
     """Stub class for server connection"""
 
-    def __init__(self, peer=None, public_key_auth=False, override_pk_ok=False,
+    def __init__(self, peer=None, gss_host=None, override_gss_mech=False,
+                 public_key_auth=False, override_pk_ok=False,
                  password_auth=False, password_change_prompt=None,
                  kbdint_auth=False, kbdint_challenge=False, success=False):
         super().__init__(peer, True)
 
+        self._gss = GSSServer(gss_host) if gss_host else None
+        self._override_gss_mech = override_gss_mech
+
         self._public_key_auth = public_key_auth
         self._override_pk_ok = override_pk_ok
 
@@ -233,10 +252,13 @@ class _AuthServerStub(_AuthConnectionStub):
 
         self._auth = None
 
-    def connection_lost(self, exc):
+    def connection_lost(self, exc=None):
         """Handle the closing of a connection"""
 
-        self._peer.connection_lost(exc)
+        if self._peer:
+            self._peer.connection_lost(exc)
+
+        self.close()
 
     def process_packet(self, data):
         """Process an incoming packet"""
@@ -252,16 +274,16 @@ class _AuthServerStub(_AuthConnectionStub):
             if self._auth:
                 self._auth.cancel()
 
-            if self._override_pk_ok:
+            if self._override_gss_mech:
+                self.send_packet(Byte(MSG_USERAUTH_GSSAPI_RESPONSE),
+                                 String('mismatch'))
+            elif self._override_pk_ok:
                 self.send_packet(Byte(MSG_USERAUTH_PK_OK),
                                  String(''), String(''))
             else:
                 self._auth = lookup_server_auth(self, 'user', method, packet)
         else:
-            try:
-                self._auth.process_packet(pkttype, packet)
-            except DisconnectError as exc:
-                self.connection_lost(exc)
+            self._auth.process_packet(pkttype, packet)
 
     def send_userauth_failure(self, partial_success):
         """Send a user authentication failure response"""
@@ -276,6 +298,31 @@ class _AuthServerStub(_AuthConnectionStub):
         self._auth = None
         self.send_packet(Byte(MSG_USERAUTH_SUCCESS))
 
+    def get_gss_context(self):
+        """Return the GSS context associated with this connection"""
+
+        return self._gss
+
+    def gss_kex_auth_supported(self):
+        """Return whether or not GSS key exchange authentication is supported"""
+
+        # pylint: disable=no-self-use
+
+        return bool(self._gss)
+
+    def gss_mic_auth_supported(self):
+        """Return whether or not GSS MIC authentication is supported"""
+
+        return bool(self._gss)
+
+    @asyncio.coroutine
+    def validate_gss_principal(self, username, user_principal, host_principal):
+        """Validate the GSS principal name for the specified user"""
+
+        # pylint: disable=unused-argument
+
+        return self._success
+
     def public_key_auth_supported(self):
         """Return whether or not public key authentication is supported"""
 
@@ -339,6 +386,7 @@ class _AuthServerStub(_AuthConnectionStub):
         return self._success
 
 
+@patch_gss
 class _TestAuth(AsyncTestCase):
     """Unit tests for auth module"""
 
@@ -373,12 +421,19 @@ class _TestAuth(AsyncTestCase):
             server_conn.close()
 
         with self.subTest('All auth methods'):
-            server_conn = _AuthServerStub(public_key_auth=True,
-                                          password_auth=True,
-                                          kbdint_auth=True)
-            self.assertEqual(get_server_auth_methods(server_conn),
-                             [b'publickey', b'keyboard-interactive',
-                              b'password'])
+            gss_host = '1' if gss_available else None
+            server_conn = _AuthServerStub(gss_host=gss_host,
+                                          public_key_auth=True,
+                                          password_auth=True, kbdint_auth=True)
+            if gss_available: # pragma: no branch
+                self.assertEqual(get_server_auth_methods(server_conn),
+                                 [b'gssapi-keyex', b'gssapi-with-mic',
+                                  b'publickey', b'keyboard-interactive',
+                                  b'password'])
+            else: # pragma: no cover
+                self.assertEqual(get_server_auth_methods(server_conn),
+                                 [b'publickey', b'keyboard-interactive',
+                                  b'password'])
             server_conn.close()
 
         with self.subTest('Unknown auth method'):
@@ -393,6 +448,51 @@ class _TestAuth(AsyncTestCase):
 
         yield from self.check_auth(b'none', (False, None))
 
+    @unittest.skipUnless(gss_available, 'GSS not available')
+    @asynctest
+    def test_gss_auth(self):
+        """Unit test GSS authentication"""
+
+        with self.subTest('GSS with MIC auth not available'):
+            yield from self.check_auth(b'gssapi-with-mic', (False, None))
+
+        for steps in range(4):
+            with self.subTest('GSS with MIC auth available'):
+                yield from self.check_auth(b'gssapi-with-mic', (True, None),
+                                           gss_host=str(steps), success=True)
+
+            gss_host = str(steps) + ',step_error'
+
+            with self.subTest('GSS with MIC error', steps=steps):
+                yield from self.check_auth(b'gssapi-with-mic', (False, None),
+                                           gss_host=gss_host)
+
+            with self.subTest('GSS with MIC error with token', steps=steps):
+                yield from self.check_auth(b'gssapi-with-mic', (False, None),
+                                           gss_host=gss_host + ',errtok')
+
+        with self.subTest('GSS with MIC without integrity'):
+            yield from self.check_auth(b'gssapi-with-mic', (True, None),
+                                       gss_host='1,no_client_integrity,' +
+                                       'no_server_integrity', success=True)
+
+        with self.subTest('GSS client integrity mismatch'):
+            yield from self.check_auth(b'gssapi-with-mic', (False, None),
+                                       gss_host='1,no_client_integrity')
+
+        with self.subTest('GSS server integrity mismatch'):
+            yield from self.check_auth(b'gssapi-with-mic', (False, None),
+                                       gss_host='1,no_server_integrity')
+
+        with self.subTest('GSS mechanism unknown'):
+            yield from self.check_auth(b'gssapi-with-mic', (False, None),
+                                       gss_host='1,unknown_mech')
+
+        with self.subTest('GSS mechanism mismatch'):
+            with self.assertRaises(DisconnectError):
+                yield from self.check_auth(b'gssapi-with-mic', (False, None),
+                                           gss_host='1', override_gss_mech=True)
+
     @asynctest
     def test_publickey_auth(self):
         """Unit test public key authentication"""
@@ -405,21 +505,23 @@ class _TestAuth(AsyncTestCase):
 
         with self.subTest('Untrusted key'):
             yield from self.check_auth(b'publickey', (False, None),
-                                       client_key=ckey)
+                                       client_key=ckey, public_key_auth=True)
 
         with self.subTest('Trusted key'):
             yield from self.check_auth(b'publickey', (True, None),
-                                       client_key=ckey, success=True)
+                                       client_key=ckey, public_key_auth=True,
+                                       success=True)
 
         with self.subTest('Trusted certificate'):
             yield from self.check_auth(b'publickey', (True, None),
                                        client_key=ckey, client_cert=cert,
-                                       success=True)
+                                       public_key_auth=True, success=True)
 
         with self.subTest('Invalid PK_OK message'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'publickey', (False, None),
                                            client_key=ckey,
+                                           public_key_auth=True,
                                            override_pk_ok=True)
 
     @asynctest
@@ -432,37 +534,40 @@ class _TestAuth(AsyncTestCase):
         with self.subTest('Invalid password'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'password', (False, None),
+                                           password_auth=True,
                                            password=b'\xff')
 
         with self.subTest('Incorrect password'):
             yield from self.check_auth(b'password', (False, None),
-                                       password='password')
+                                       password_auth=True, password='password')
 
         with self.subTest('Correct password'):
             yield from self.check_auth(b'password', (True, None),
-                                       password='password', success=True)
+                                       password_auth=True, password='password',
+                                       success=True)
 
         with self.subTest('Password change not available'):
             yield from self.check_auth(b'password', (False, None),
-                                       password='password',
+                                       password_auth=True, password='password',
                                        password_change_prompt='change')
 
         with self.subTest('Invalid password change prompt'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'password', (False, False),
+                                           password_auth=True,
                                            password='password',
                                            password_change=True,
                                            password_change_prompt=b'\xff')
 
         with self.subTest('Password change failed'):
             yield from self.check_auth(b'password', (False, False),
-                                       password='password',
+                                       password_auth=True, password='password',
                                        password_change=True,
                                        password_change_prompt='change')
 
         with self.subTest('Password change succeeded'):
             yield from self.check_auth(b'password', (True, True),
-                                       password='password',
+                                       password_auth=True, password='password',
                                        password_change=True,
                                        password_change_prompt='change',
                                        success=True)
@@ -471,54 +576,61 @@ class _TestAuth(AsyncTestCase):
     def test_kbdint_auth(self):
         """Unit test keyboard-interactive authentication"""
 
-        with self.subTest('No submethods'):
+        with self.subTest('Kbdint auth not available'):
             yield from self.check_auth(b'keyboard-interactive', (False, None))
 
+        with self.subTest('No submethods'):
+            yield from self.check_auth(b'keyboard-interactive', (False, None),
+                                       kbdint_auth=True)
+
         with self.subTest('Invalid submethods'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'keyboard-interactive',
-                                           (False, None),
+                                           (False, None), kbdint_auth=True,
                                            kbdint_submethods=b'\xff')
 
         with self.subTest('No challenge'):
             yield from self.check_auth(b'keyboard-interactive', (False, None),
-                                       kbdint_submethods='')
+                                       kbdint_auth=True, kbdint_submethods='')
 
         with self.subTest('Invalid challenge name'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'keyboard-interactive',
-                                           (False, None), kbdint_submethods='',
+                                           (False, None), kbdint_auth=True,
+                                           kbdint_submethods='',
                                            kbdint_challenge=(b'\xff', '',
                                                              '', ()))
 
         with self.subTest('Invalid challenge prompt'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'keyboard-interactive',
-                                           (False, None), kbdint_submethods='',
+                                           (False, None), kbdint_auth=True,
+                                           kbdint_submethods='',
                                            kbdint_challenge=('', '', '',
                                                              ((b'\xff',
                                                                False),)))
 
         with self.subTest('No response'):
             yield from self.check_auth(b'keyboard-interactive', (False, None),
-                                       kbdint_submethods='',
+                                       kbdint_auth=True, kbdint_submethods='',
                                        kbdint_challenge=True)
 
         with self.subTest('Invalid response'):
             with self.assertRaises(DisconnectError):
                 yield from self.check_auth(b'keyboard-interactive',
-                                           (False, None), kbdint_submethods='',
+                                           (False, None), kbdint_auth=True,
+                                           kbdint_submethods='',
                                            kbdint_challenge=True,
                                            kbdint_response=(b'\xff',))
 
         with self.subTest('Incorrect response'):
             yield from self.check_auth(b'keyboard-interactive', (False, None),
-                                       kbdint_submethods='',
+                                       kbdint_auth=True, kbdint_submethods='',
                                        kbdint_challenge=True,
                                        kbdint_response=True)
 
         with self.subTest('Correct response'):
             yield from self.check_auth(b'keyboard-interactive', (True, None),
-                                       kbdint_submethods='',
+                                       kbdint_auth=True, kbdint_submethods='',
                                        kbdint_challenge=True,
                                        kbdint_response=True, success=True)
