@@ -13,12 +13,12 @@
 """Unit tests for AsyncSSH channel API"""
 
 import asyncio
+import tempfile
 
 from unittest.mock import patch
 
 import asyncssh
 
-from asyncssh.channel import SSHAgentChannel, SSHClientChannel
 from asyncssh.constants import DEFAULT_LANG, MSG_USERAUTH_REQUEST
 from asyncssh.constants import MSG_CHANNEL_OPEN_CONFIRMATION
 from asyncssh.constants import MSG_CHANNEL_OPEN_FAILURE
@@ -29,7 +29,6 @@ from asyncssh.constants import MSG_CHANNEL_EOF, MSG_CHANNEL_CLOSE
 from asyncssh.constants import MSG_CHANNEL_SUCCESS
 from asyncssh.packet import Byte, String, UInt32
 from asyncssh.public_key import CERT_TYPE_USER
-from asyncssh.stream import SSHClientStreamSession
 from asyncssh.stream import SSHTCPStreamSession, SSHUNIXStreamSession
 
 from .server import Server, ServerTestCase
@@ -54,6 +53,11 @@ class _ClientChannel(asyncssh.SSHClientChannel):
 
         super()._send_request(request, *args, want_reply=want_reply)
 
+    def send_request(self, request, *args):
+        """Send a custom request (for unit testing)"""
+
+        self._send_request(request, *args)
+
     @asyncio.coroutine
     def make_request(self, request, *args):
         """Make a custom request (for unit testing)"""
@@ -132,11 +136,24 @@ class _ServerChannel(asyncssh.SSHServerChannel):
 
         super()._send_request(request, *args, want_reply=want_reply)
 
+    def _process_delayed_request(self, packet):
+        """Process a request that delays before responding"""
+
+        packet.check_end()
+
+        asyncio.get_event_loop().call_later(0.1, self._report_response, True)
+
     def send_packet(self, pkttype, *args):
         """Send a packet for unit testing (bypassing state checks)"""
 
         self._send_packet(pkttype, *args)
 
+    @asyncio.coroutine
+    def open_session(self):
+        """Attempt to open a session on the client"""
+
+        return (yield from self._open(b'session'))
+
 
 class _EchoServerSession(asyncssh.SSHServerSession):
     """A shell session which echos data from stdin to stdout/stderr"""
@@ -212,21 +229,31 @@ class _ChannelServer(Server):
                 agent.close()
             else:
                 stdout.channel.exit(1)
+        elif action == 'agent_sock':
+            agent_path = stdin.channel.get_agent_path()
+
+            if agent_path:
+                agent = yield from asyncssh.connect_agent(agent_path)
+                stdout.write(str(len((yield from agent.get_keys()))) + '\n')
+                agent.close()
+            else:
+                stdout.channel.exit(1)
         elif action == 'rejected_agent':
-            chan = SSHAgentChannel(self._conn, asyncio.get_event_loop(),
-                                   None, 1, 32768)
+            agent_path = stdin.channel.get_agent_path()
+            stdout.write(str(bool(agent_path)) + '\n')
+
+            chan = self._conn.create_agent_channel()
 
             try:
                 yield from chan.open(SSHUNIXStreamSession)
             except asyncssh.ChannelOpenError:
                 stdout.channel.exit(1)
         elif action == 'rejected_session':
-            chan = SSHClientChannel(self._conn, asyncio.get_event_loop(),
-                                    None, 1, 32768)
+            chan = _ServerChannel(self._conn, asyncio.get_event_loop(),
+                                  False, False, 0, None, 1, 32768)
 
             try:
-                yield from chan.create(SSHClientStreamSession, None, None,
-                                       {}, None, None, None, False)
+                yield from chan.open_session()
             except asyncssh.ChannelOpenError:
                 stdout.channel.exit(1)
         elif action == 'rejected_tcpip_direct':
@@ -649,6 +676,19 @@ class _TestChannel(ServerTestCase):
 
             yield from conn.wait_closed()
 
+    @asynctest
+    def test_delayed_channel_request(self):
+        """Test queuing channel requests with delayed response"""
+
+        with patch('asyncssh.connection.SSHClientChannel', _ClientChannel):
+            with (yield from self.connect()) as conn:
+                chan, _ = yield from _create_session(conn)
+
+                chan.send_request(b'delayed')
+                chan.send_request(b'delayed')
+
+            yield from conn.wait_closed()
+
     @asynctest
     def test_invalid_channel_response(self):
         """Test receiving response for non-existent channel request"""
@@ -669,7 +709,7 @@ class _TestChannel(ServerTestCase):
 
             with self.assertRaises(OSError):
                 yield from chan.create(None, None, None, {}, None, None,
-                                       None, False)
+                                       None, False, None, None, False, False)
 
             chan.close()
 
@@ -746,6 +786,31 @@ class _TestChannel(ServerTestCase):
             result = ''.join(session.recv_buf[None])
             self.assertEqual(result, '3\n')
 
+            chan, session = yield from _create_session(conn, 'agent')
+
+            yield from chan.wait_closed()
+
+            result = ''.join(session.recv_buf[None])
+            self.assertEqual(result, '3\n')
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_agent_forwarding_sock(self):
+        """Test SSH agent forwarding via UNIX domain socket"""
+
+        if not self.agent_available(): # pragma: no cover
+            self.skipTest('ssh-agent not available')
+
+        with (yield from self.connect(username='ckey',
+                                      agent_forwarding=True)) as conn:
+            chan, session = yield from _create_session(conn, 'agent_sock')
+
+            yield from chan.wait_closed()
+
+            result = ''.join(session.recv_buf[None])
+            self.assertEqual(result, '3\n')
+
         yield from conn.wait_closed()
 
     @asynctest
@@ -810,6 +875,24 @@ class _TestChannel(ServerTestCase):
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_agent_forwarding_sock_failure(self):
+        """Test failure to create SSH agent forwarding socket"""
+
+        tempfile.tempdir = 'xxx'
+
+        with (yield from self.connect(username='ckey',
+                                      agent_forwarding=True)) as conn:
+            chan, session = yield from _create_session(conn, 'agent_sock')
+
+            yield from chan.wait_closed()
+
+            self.assertEqual(session.exit_status, 1)
+
+        yield from conn.wait_closed()
+
+        tempfile.tempdir = None
+
     @asynctest
     def test_agent_forwarding_not_offered(self):
         """Test SSH agent forwarding not offered by client"""
@@ -832,6 +915,9 @@ class _TestChannel(ServerTestCase):
 
             yield from chan.wait_closed()
 
+            result = ''.join(session.recv_buf[None])
+            self.assertEqual(result, 'False\n')
+
             self.assertEqual(session.exit_status, 1)
 
         yield from conn.wait_closed()
