@@ -1,4 +1,4 @@
-# Copyright (c) 2016 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2016-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -13,9 +13,9 @@
 """Unit tests for AsyncSSH connection API"""
 
 import asyncio
-import os
-
 from copy import copy
+import os
+import unittest
 from unittest.mock import patch
 
 import asyncssh
@@ -35,7 +35,7 @@ from asyncssh.mac import _HMAC, _mac_handlers, get_mac_algs
 from asyncssh.packet import Boolean, Byte, NameList, String, UInt32
 
 from .server import Server, ServerTestCase
-from .util import asynctest
+from .util import asynctest, gss_available, patch_gss, x509_available
 
 
 class _SplitClientConnection(asyncssh.SSHClientConnection):
@@ -151,6 +151,24 @@ class _AbortServer(Server):
         return False
 
 
+class _CloseDuringAuthServer(Server):
+    """Server for testing connection close during long auth callback"""
+
+    def password_auth_supported(self):
+        """Return that password auth is supported"""
+
+        return True
+
+    @asyncio.coroutine
+    def validate_password(self, username, password):
+        """Delay validating password"""
+
+        # pylint: disable=unused-argument
+
+        yield from asyncio.sleep(1)
+        return False
+
+
 class _InternalErrorServer(Server):
     """Server for testing internal error during auth"""
 
@@ -193,11 +211,19 @@ class _VersionReportingServer(Server):
         return False
 
 
+@patch_gss
 class _TestConnection(ServerTestCase):
     """Unit tests for AsyncSSH connection API"""
 
     # pylint: disable=too-many-public-methods
 
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server to connect to"""
+
+        return (yield from cls.create_server(gss_host=()))
+
     @asyncio.coroutine
     def _check_version(self, *args, **kwargs):
         """Check alternate SSH server version lines"""
@@ -273,7 +299,7 @@ class _TestConnection(ServerTestCase):
         """Test starting a server with no host keys"""
 
         with self.assertRaises(ValueError):
-            yield from asyncssh.listen(server_host_keys=[])
+            yield from asyncssh.listen(server_host_keys=[], gss_host=None)
 
     @asynctest
     def test_duplicate_type_server_host_keys(self):
@@ -291,6 +317,16 @@ class _TestConnection(ServerTestCase):
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_known_hosts_none_without_x509(self):
+        """Test connecting with known hosts checking and X.509 disabled"""
+
+        with (yield from self.connect(known_hosts=None,
+                                      x509_trusted_certs=None)) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
     @asynctest
     def test_known_hosts_multiple_keys(self):
         """Test connecting with multiple trusted known hosts keys"""
@@ -337,13 +373,39 @@ class _TestConnection(ServerTestCase):
     def test_known_hosts_sshkeys(self):
         """Test connecting with known hosts passed in as SSHKeys"""
 
-        keylist = asyncssh.read_public_key_list('skey.pub')
+        keylist = asyncssh.load_public_keys('skey.pub')
 
         with (yield from self.connect(known_hosts=(keylist, [], []))) as conn:
             pass
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_read_known_hosts(self):
+        """Test connecting with known hosts object from read_known_hosts"""
+
+        known_hosts_path = os.path.join('.ssh', 'known_hosts')
+        known_hosts = asyncssh.read_known_hosts(known_hosts_path)
+
+        with (yield from self.connect(known_hosts=known_hosts)) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_import_known_hosts(self):
+        """Test connecting with known hosts object from import_known_hosts"""
+
+        known_hosts_path = os.path.join('.ssh', 'known_hosts')
+
+        with open(known_hosts_path, 'r') as f:
+            known_hosts = asyncssh.import_known_hosts(f.read())
+
+        with (yield from self.connect(known_hosts=known_hosts)) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
     @asynctest
     def test_untrusted_known_hosts_key(self):
         """Test untrusted server host key"""
@@ -397,8 +459,12 @@ class _TestConnection(ServerTestCase):
         for kex in get_kex_algs():
             kex = kex.decode('ascii')
 
+            if kex.startswith('gss-') and not gss_available: # pragma: no cover
+                continue
+
             with self.subTest(kex_alg=kex):
-                with (yield from self.connect(kex_algs=[kex])) as conn:
+                with (yield from self.connect(kex_algs=[kex],
+                                              gss_host='1')) as conn:
                     pass
 
                 yield from conn.wait_closed()
@@ -694,6 +760,20 @@ class _TestConnection(ServerTestCase):
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_no_matching_host_key_algs(self):
+        """Test no matching server host key algorithms"""
+
+        conn = yield from self.connect()
+
+        conn.send_packet(Byte(MSG_KEXINIT), os.urandom(16),
+                         NameList([b'ecdh-sha2-nistp521']), NameList([b'xxx']),
+                         NameList([]), NameList([]), NameList([]),
+                         NameList([]), NameList([]), NameList([]),
+                         NameList([]), NameList([]), Boolean(False), UInt32(0))
+
+        yield from conn.wait_closed()
+
     @asynctest
     def test_invalid_newkeys(self):
         """Test invalid new keys request"""
@@ -892,13 +972,178 @@ class _TestConnectionAbort(ServerTestCase):
             yield from self.connect()
 
 
+class _TestConnectionCloseDurngAuth(ServerTestCase):
+    """Unit test for connection close during long auth callback"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which aborts connections during auth"""
+
+        return (yield from cls.create_server(_CloseDuringAuthServer))
+
+    @asynctest
+    def test_close_during_auth(self):
+        """Test connection close during long auth callback"""
+
+        with self.assertRaises(asyncio.TimeoutError):
+            yield from asyncio.wait_for(self.connect(username='user',
+                                                     password=''), 0.5)
+
+
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestServerX509Self(ServerTestCase):
+    """Unit test for server with self-signed X.509 host certificate"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server with a self-signed X.509 host certificate"""
+
+        return (yield from cls.create_server(
+            server_host_keys=['skey_x509_self']))
+
+    @asynctest
+    def test_connect_x509_self(self):
+        """Test connecting with X.509 self-signed certificate"""
+
+        with (yield from self.connect(known_hosts=([], [], [],
+                                                   ['skey_x509_self.pem'],
+                                                   [], [], []))) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_connect_x509_untrusted_self(self):
+        """Test connecting with untrusted X.509 self-signed certficate"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(
+                known_hosts=([], [], [], ['root_ca_cert.pem'], [], [], []))
+
+    @asynctest
+    def test_connect_x509_revoked_self(self):
+        """Test connecting with revoked X.509 self-signed certficate"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(
+                known_hosts=([], [], [], ['root_ca_cert.pem'],
+                             ['skey_x509_self.pem'], [], []))
+
+    @asynctest
+    def test_connect_x509_trusted_subject(self):
+        """Test connecting to server with trusted X.509 subject name"""
+
+        with (yield from self.connect(
+            known_hosts=([], [], [], [], [], ['OU=name'], ['OU=name1']),
+            x509_trusted_certs=['skey_x509_self.pem'])) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_connect_x509_untrusted_subject(self):
+        """Test connecting to server with untrusted X.509 subject name"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(
+                known_hosts=([], [], [], [], [], ['OU=name1'], []),
+                x509_trusted_certs=['skey_x509_self.pem'])
+
+    @asynctest
+    def test_connect_x509_revoked_subject(self):
+        """Test connecting to server with revoked X.509 subject name"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(
+                known_hosts=([], [], [], [], [], [], ['OU=name']),
+                x509_trusted_certs=['skey_x509_self.pem'])
+
+    @asynctest
+    def test_connect_x509_disabled(self):
+        """Test connecting to X.509 server with X.509 disabled"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(
+                known_hosts=([], [], [], [], [], ['OU=name'], []),
+                x509_trusted_certs=None)
+
+
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestServerX509Chain(ServerTestCase):
+    """Unit test for server with X.509 host certificate chain"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server with an X.509 host certificate chain"""
+
+        return (yield from cls.create_server(
+            server_host_keys=['skey_x509_chain']))
+
+    @asynctest
+    def test_connect_x509_chain(self):
+        """Test connecting with X.509 certificate chain"""
+
+        with (yield from self.connect(known_hosts=([], [], [],
+                                                   ['root_ca_cert.pem'],
+                                                   [], [], []))) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_connect_x509_chain_cert_path(self):
+        """Test connecting with X.509 certificate and certificate path"""
+
+        with (yield from self.connect(x509_trusted_cert_paths=['cert_path'],
+                                      known_hosts=b'\n')) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_connect_x509_untrusted_root(self):
+        """Test connecting to server with untrusted X.509 root CA"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(known_hosts=([], [], [],
+                                                 ['skey_x509_self.pem'],
+                                                 [], [], []))
+
+    @asynctest
+    def test_connect_x509_untrusted_root_cert_path(self):
+        """Test connecting to server with untrusted X.509 root CA"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(known_hosts=b'\n')
+
+    @asynctest
+    def test_connect_x509_revoked_intermediate(self):
+        """Test connecting to server with revoked X.509 intermediate CA"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(known_hosts=([], [], [],
+                                                 ['root_ca_cert.pem'],
+                                                 ['int_ca_cert.pem'],
+                                                 [], []))
+
+    @asynctest
+    def test_invalid_x509_path(self):
+        """Test passing in invalid trusted X.509 certificate path"""
+
+        with self.assertRaises(ValueError):
+            yield from self.connect(x509_trusted_cert_paths='xxx')
+
+
 class _TestServerNoLoop(ServerTestCase):
     """Unit test for server with no loop specified"""
 
     @classmethod
     @asyncio.coroutine
     def start_server(cls):
-        """Start an SSH server which raises an error during auth"""
+        """Start an SSH server which has no loop specified"""
 
         return (yield from cls.create_server(loop=None))
 
@@ -912,6 +1157,38 @@ class _TestServerNoLoop(ServerTestCase):
         yield from conn.wait_closed()
 
 
+@unittest.skipUnless(gss_available, 'GSS not available')
+@patch_gss
+class _TestServerNoHostKey(ServerTestCase):
+    """Unit test for server with no server host key"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which sets no server host keys"""
+
+        return (yield from cls.create_server(server_host_keys=None,
+                                             gss_host='1'))
+
+    @asynctest
+    def test_gss_with_no_host_key(self):
+        """Test GSS key exchange with no server host key specified"""
+
+        with (yield from self.connect(known_hosts=b'\n', gss_host='1',
+                                      x509_trusted_certs=None,
+                                      x509_trusted_cert_paths=None)) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_dh_with_no_host_key(self):
+        """Test failure of DH key exchange with no server host key specified"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect()
+
+
 class _TestServerInternalError(ServerTestCase):
     """Unit test for server internal error during auth"""
 
