@@ -1,4 +1,4 @@
-# Copyright (c) 2016 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2016-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -14,6 +14,7 @@
 
 import asyncio
 import os
+import unittest
 
 from unittest.mock import patch
 
@@ -22,7 +23,19 @@ from asyncssh.packet import String
 from asyncssh.public_key import CERT_TYPE_USER
 
 from .server import Server, ServerTestCase
-from .util import asynctest, make_certificate
+from .util import asynctest, gss_available, patch_gss, make_certificate
+from .util import x509_available
+
+
+class _AsyncGSSServer(asyncssh.SSHServer):
+    """Server for testing async GSS authentication"""
+
+    @asyncio.coroutine
+    def validate_gss_principal(self, username, user_principal, host_principal):
+        """Return whether password is valid for this user"""
+
+        return super().validate_gss_principal(username, user_principal,
+                                              host_principal)
 
 
 class _PublicKeyClient(asyncssh.SSHClient):
@@ -55,9 +68,9 @@ class _AsyncPublicKeyClient(_PublicKeyClient):
 class _PublicKeyServer(Server):
     """Server for testing public key authentication"""
 
-    def __init__(self, authorized_keys=None):
+    def __init__(self, client_keys=(), authorized_keys=None):
         super().__init__()
-        self._client_keys = []
+        self._client_keys = client_keys
         self._authorized_keys = authorized_keys
 
     def connection_made(self, conn):
@@ -72,8 +85,7 @@ class _PublicKeyServer(Server):
         if self._authorized_keys:
             self._conn.set_authorized_keys(self._authorized_keys)
         else:
-            self._client_keys = asyncssh.load_public_keys(['ckey_dsa.pub',
-                                                           'ckey.pub'])
+            self._client_keys = asyncssh.load_public_keys(self._client_keys)
 
         return True
 
@@ -289,6 +301,107 @@ class _UnknownAuthClientConnection(asyncssh.connection.SSHClientConnection):
         super().try_next_auth()
 
 
+@unittest.skipUnless(gss_available, 'GSS not available')
+@patch_gss
+class _TestGSSAuth(ServerTestCase):
+    """Unit tests for GSS authentication"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which supports GSS authentication"""
+
+        return (yield from cls.create_server(_AsyncGSSServer, gss_host='1'))
+
+    @asynctest
+    def test_gss_kex_auth(self):
+        """Test GSS key exchange authentication"""
+
+        with (yield from self.connect(kex_algs=['gss-gex-sha1'],
+                                      username='user', gss_host='1')) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_gss_mic_auth(self):
+        """Test GSS MIC authentication"""
+
+        with (yield from self.connect(kex_algs=['ecdh-sha2-nistp256'],
+                                      username='user', gss_host='1')) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_gss_auth_unavailable(self):
+        """Test GSS authentication being unavailable"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='user1', gss_host=())
+
+    @asynctest
+    def test_gss_client_error(self):
+        """Test GSS client error"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(gss_host='1,init_error', username='user')
+
+
+@unittest.skipUnless(gss_available, 'GSS not available')
+@patch_gss
+class _TestGSSServerError(ServerTestCase):
+    """Unit tests for GSS server error"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which raises an error on GSS authentication"""
+
+        return (yield from cls.create_server(gss_host='1,init_error'))
+
+    @asynctest
+    def test_gss_server_error(self):
+        """Test GSS error on server"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='user')
+
+
+@unittest.skipUnless(gss_available, 'GSS not available')
+@patch_gss
+class _TestGSSFQDN(ServerTestCase):
+    """Unit tests for GSS server error"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which raises an error on GSS authentication"""
+
+        def mock_gethostname():
+            """Return a non-fully-qualified hostname"""
+
+            return 'host'
+
+        def mock_getfqdn():
+            """Confirm getfqdn is called on relative hostnames"""
+
+            return '1'
+
+        with patch('socket.gethostname', mock_gethostname):
+            with patch('socket.getfqdn', mock_getfqdn):
+                return (yield from cls.create_server(gss_host=()))
+
+    @asynctest
+    def test_gss_fqdn_lookup(self):
+        """Test GSS FQDN lookup"""
+
+        with (yield from self.connect(username='user', gss_host=())) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+
 class _TestPublicKeyAuth(ServerTestCase):
     """Unit tests for public key authentication"""
 
@@ -301,14 +414,14 @@ class _TestPublicKeyAuth(ServerTestCase):
             _PublicKeyServer, authorized_client_keys='authorized_keys'))
 
     @asyncio.coroutine
-    def _connect_publickey(self, keylist, async=False):
+    def _connect_publickey(self, keylist, test_async=False):
         """Open a connection to test public key auth"""
 
         def client_factory():
             """Return an SSHClient to use to do public key auth"""
 
-            client_class = _AsyncPublicKeyClient if async else _PublicKeyClient
-            return client_class(keylist)
+            cls = _AsyncPublicKeyClient if test_async else _PublicKeyClient
+            return cls(keylist)
 
         conn, _ = yield from self.create_connection(client_factory,
                                                     username='ckey',
@@ -413,6 +526,19 @@ class _TestPublicKeyAuth(ServerTestCase):
 
         yield from conn.wait_closed()
 
+    @asynctest
+    def test_invalid_default_key(self):
+        """Test connecting with invalid default client key"""
+
+        key_path = os.path.join('.ssh', 'id_dsa')
+        with open(key_path, 'w') as f:
+            f.write('-----XXX-----')
+
+        with self.assertRaises(asyncssh.KeyImportError):
+            yield from self.connect(username='ckey', agent_path=None)
+
+        os.remove(key_path)
+
     @asynctest
     def test_client_key_bytes(self):
         """Test client key passed in as bytes"""
@@ -495,8 +621,7 @@ class _TestPublicKeyAuth(ServerTestCase):
                                 valid_before=1)
 
         with self.assertRaises(asyncssh.DisconnectError):
-            yield from self.connect(username='ckey',
-                                    client_keys=[(skey, cert)])
+            yield from self.connect(username='ckey', client_keys=[(skey, cert)])
 
     @asynctest
     def test_allowed_address(self):
@@ -528,8 +653,7 @@ class _TestPublicKeyAuth(ServerTestCase):
                                 options={'source-address': String('0.0.0.0')})
 
         with self.assertRaises(asyncssh.DisconnectError):
-            yield from self.connect(username='ckey',
-                                    client_keys=[(skey, cert)])
+            yield from self.connect(username='ckey', client_keys=[(skey, cert)])
 
     @asynctest
     def test_untrusted_ca(self):
@@ -549,7 +673,7 @@ class _TestPublicKeyAuth(ServerTestCase):
         """Test connecting with public key authentication using callback"""
 
         with (yield from self._connect_publickey(['ckey'],
-                                                 async=True)) as conn:
+                                                 test_async=True)) as conn:
             pass
 
         yield from conn.wait_closed()
@@ -599,7 +723,13 @@ class _TestPublicKeyAsyncServerAuth(_TestPublicKeyAuth):
     def start_server(cls):
         """Start an SSH server which supports async public key auth"""
 
-        return (yield from cls.create_server(_AsyncPublicKeyServer))
+        def server_factory():
+            """Return an SSH server which calls set_authorized_keys"""
+
+            return _AsyncPublicKeyServer(client_keys=['ckey.pub',
+                                                      'ckey_ecdsa.pub'])
+
+        return (yield from cls.create_server(server_factory))
 
 
 class _TestLimitedSignatureAlgs(ServerTestCase):
@@ -702,6 +832,157 @@ class _TestPreloadedAuthorizedKeys(ServerTestCase):
         yield from conn.wait_closed()
 
 
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestX509Auth(ServerTestCase):
+    """Unit tests for X.509 certificate authentication"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which supports public key authentication"""
+
+        return (yield from cls.create_server(
+            _PublicKeyServer, authorized_client_keys='authorized_keys_x509'))
+
+    @asynctest
+    def test_x509_self(self):
+        """Test connecting with X.509 self-signed certificate"""
+
+        with (yield from self.connect(username='ckey',
+                                      client_keys=['ckey_x509_self'])) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_x509_chain(self):
+        """Test connecting with X.509 certificate chain"""
+
+        with (yield from self.connect(username='ckey',
+                                      client_keys=['ckey_x509_chain'])) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+    @asynctest
+    def test_x509_incomplete_chain(self):
+        """Test connecting with incomplete X.509 certificate chain"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='ckey',
+                                    client_keys=[('ckey_x509_chain',
+                                                  'ckey_x509_partial.pem')])
+
+    @asynctest
+    def test_x509_untrusted_cert(self):
+        """Test connecting with untrusted X.509 certificate chain"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='ckey',
+                                    client_keys=['skey_x509_chain'])
+
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestX509AuthDisabled(ServerTestCase):
+    """Unit tests for disabled X.509 certificate authentication"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which doesn't support X.509 authentication"""
+
+        return (yield from cls.create_server(
+            _PublicKeyServer, x509_trusted_certs=None,
+            authorized_client_keys='authorized_keys'))
+
+    @asynctest
+    def test_failed_x509_auth(self):
+        """Test connect failure with X.509 certificate"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='ckey',
+                                    client_keys=['ckey_x509_self'],
+                                    signature_algs=['x509v3-ssh-rsa'])
+
+    @asynctest
+    def test_non_x509(self):
+        """Test connecting without an X.509 certificate"""
+
+        with (yield from self.connect(username='ckey',
+                                      client_keys=['ckey'])) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestX509Subject(ServerTestCase):
+    """Unit tests for X.509 certificate authentication by subject name"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which supports public key authentication"""
+
+        authorized_keys = asyncssh.import_authorized_keys(
+            'x509v3-ssh-rsa subject=OU=name\n')
+
+        return (yield from cls.create_server(
+            _PublicKeyServer, authorized_client_keys=authorized_keys,
+            x509_trusted_certs=['ckey_x509_self.pub']))
+
+    @asynctest
+    def test_x509_subject(self):
+        """Test authenticating X.509 certificate by subject name"""
+
+        with (yield from self.connect(username='ckey',
+                                      client_keys=['ckey_x509_self'])) as conn:
+            pass
+
+        yield from conn.wait_closed()
+
+
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestX509Untrusted(ServerTestCase):
+    """Unit tests for X.509 authentication with no trusted certificates"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server which supports public key authentication"""
+
+        return (yield from cls.create_server(
+            _PublicKeyServer, authorized_client_keys=None))
+
+    @asynctest
+    def test_x509_untrusted(self):
+        """Test untrusted X.509 self-signed certificate"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='ckey',
+                                    client_keys=['ckey_x509_self'])
+
+
+@unittest.skipUnless(x509_available, 'X.509 not available')
+class _TestX509Disabled(ServerTestCase):
+    """Unit tests for X.509 authentication with server support disabled"""
+
+    @classmethod
+    @asyncio.coroutine
+    def start_server(cls):
+        """Start an SSH server with X.509 authentication disabled"""
+
+        return (yield from cls.create_server(_PublicKeyServer,
+                                             x509_purposes=None))
+
+    @asynctest
+    def test_x509_disabled(self):
+        """Test X.509 client certificate with server support disabled"""
+
+        with self.assertRaises(asyncssh.DisconnectError):
+            yield from self.connect(username='ckey',
+                                    client_keys='skey_x509_self')
+
+
 class _TestPasswordAuth(ServerTestCase):
     """Unit tests for password authentication"""
 
@@ -714,14 +995,14 @@ class _TestPasswordAuth(ServerTestCase):
 
     @asyncio.coroutine
     def _connect_password(self, username, password, old_password='',
-                          new_password='', async=False):
+                          new_password='', test_async=False):
         """Open a connection to test password authentication"""
 
         def client_factory():
             """Return an SSHClient to use to do password change"""
 
-            client_class = _AsyncPasswordClient if async else _PasswordClient
-            return client_class(password, old_password, new_password)
+            cls = _AsyncPasswordClient if test_async else _PasswordClient
+            return cls(password, old_password, new_password)
 
         conn, _ = yield from self.create_connection(client_factory,
                                                     username=username,
@@ -752,7 +1033,7 @@ class _TestPasswordAuth(ServerTestCase):
         """Test connecting with password authentication callback"""
 
         with (yield from self._connect_password('pw', 'pw',
-                                                async=True)) as conn:
+                                                test_async=True)) as conn:
             pass
 
         yield from conn.wait_closed()
@@ -768,8 +1049,8 @@ class _TestPasswordAuth(ServerTestCase):
     def test_password_change(self):
         """Test password change"""
 
-        with (yield from self._connect_password('pw', 'oldpw', 'oldpw',
-                                                'pw', async=True)) as conn:
+        with (yield from self._connect_password('pw', 'oldpw', 'oldpw', 'pw',
+                                                test_async=True)) as conn:
             pass
 
         yield from conn.wait_closed()
@@ -804,14 +1085,14 @@ class _TestKbdintAuth(ServerTestCase):
         return (yield from cls.create_server(_KbdintServer))
 
     @asyncio.coroutine
-    def _connect_kbdint(self, username, responses, async=False):
+    def _connect_kbdint(self, username, responses, test_async=False):
         """Open a connection to test keyboard-interactive auth"""
 
         def client_factory():
             """Return an SSHClient to use to do keyboard-interactive auth"""
 
-            client_class = _AsyncKbdintClient if async else _KbdintClient
-            return client_class(responses)
+            cls = _AsyncKbdintClient if test_async else _KbdintClient
+            return cls(responses)
 
         conn, _ = yield from self.create_connection(client_factory,
                                                     username=username,
@@ -842,7 +1123,7 @@ class _TestKbdintAuth(ServerTestCase):
         """Test keyboard-interactive auth callback"""
 
         with (yield from self._connect_kbdint('kbdint', ['kbdint'],
-                                              async=True)) as conn:
+                                              test_async=True)) as conn:
             pass
 
         yield from conn.wait_closed()
@@ -923,8 +1204,9 @@ class _TestLoginTimeoutExceeded(ServerTestCase):
     def start_server(cls):
         """Start an SSH server with a 1 second login timeout"""
 
-        return (yield from cls.create_server(_PublicKeyServer,
-                                             login_timeout=1))
+        return (yield from cls.create_server(
+            _PublicKeyServer, authorized_client_keys='authorized_keys',
+            login_timeout=1))
 
     @asynctest
     def test_login_timeout_exceeded(self):
@@ -948,8 +1230,9 @@ class _TestLoginTimeoutDisabled(ServerTestCase):
     def start_server(cls):
         """Start an SSH server with no login timeout"""
 
-        return (yield from cls.create_server(_PublicKeyServer,
-                                             login_timeout=None))
+        return (yield from cls.create_server(
+            _PublicKeyServer, authorized_client_keys='authorized_keys',
+            login_timeout=None))
 
     @asynctest
     def test_login_timeout_disabled(self):
