@@ -13,8 +13,11 @@
 """Unit tests for AsyncSSH process API"""
 
 import asyncio
+import io
 import os
 import socket
+import sys
+import unittest
 
 import asyncssh
 
@@ -22,53 +25,74 @@ from .server import ServerTestCase
 from .util import asynctest, echo
 
 
-def _handle_session(stdin, stdout, stderr):
-    """Handle a new session"""
+def _handle_client(process):
+    """Handle a new client request"""
 
     # pylint: disable=no-self-use
 
-    action = stdin.channel.get_command() or stdin.channel.get_subsystem()
+    action = process.command or process.subsystem
     if not action:
         action = 'echo'
 
     if action == 'break':
         try:
-            yield from stdin.readline()
+            yield from process.stdin.readline()
         except asyncssh.BreakReceived as exc:
-            stdin.channel.exit_with_signal('ABRT', False, str(exc.msec))
+            process.exit_with_signal('ABRT', False, str(exc.msec))
     elif action == 'delay':
         yield from asyncio.sleep(1)
-        yield from echo(stdin, stdout, stderr)
+        yield from echo(process.stdin, process.stdout, process.stderr)
     elif action == 'echo':
-        yield from echo(stdin, stdout, stderr)
+        yield from echo(process.stdin, process.stdout, process.stderr)
     elif action == 'exit_status':
-        stdin.channel.exit(1)
+        process.channel.set_encoding('utf-8')
+        process.stderr.write('Exiting with status 1')
+        process.exit(1)
+    elif action == 'env':
+        process.channel.set_encoding('utf-8')
+        process.stdout.write(process.env.get('TEST', ''))
+    elif action == 'redirect_stdin':
+        yield from process.redirect_stdin(process.stdout)
+        yield from process.stdout.drain()
+    elif action == 'redirect_stdout':
+        yield from process.redirect_stdout(process.stdin)
+        yield from process.stdout.drain()
+    elif action == 'redirect_stderr':
+        yield from process.redirect_stderr(process.stdin)
+        yield from process.stderr.drain()
     elif action == 'term':
+        info = str((process.get_terminal_type(), process.get_terminal_size(),
+                    process.get_terminal_mode(asyncssh.PTY_OP_OSPEED)))
+        process.channel.set_encoding('utf-8')
+        process.stdout.write(info)
+    elif action == 'term_size':
         try:
-            yield from stdin.readline()
+            yield from process.stdin.readline()
         except asyncssh.TerminalSizeChanged as exc:
-            stdin.channel.exit_with_signal('ABRT', False,
-                                           '%sx%s' % (exc.width, exc.height))
+            process.exit_with_signal('ABRT', False,
+                                     '%sx%s' % (exc.width, exc.height))
     else:
-        stdin.channel.exit(255)
+        process.exit(255)
 
-    stdin.channel.close()
-    yield from stdin.channel.wait_closed()
+    process.close()
+    yield from process.wait_closed()
 
 
 class _TestProcess(ServerTestCase):
     """Unit tests for AsyncSSH process API"""
 
-    # pylint: disable=too-many-public-methods
-
     @classmethod
     @asyncio.coroutine
     def start_server(cls):
         """Start an SSH server for the tests to use"""
 
-        return (yield from cls.create_server(session_factory=_handle_session,
+        return (yield from cls.create_server(process_factory=_handle_client,
                                              session_encoding=None))
 
+
+class _TestProcessBasic(_TestProcess):
+    """Unit tests for AsyncSSH process basic functions"""
+
     @asynctest
     def test_shell(self):
         """Test starting a remote shell"""
@@ -76,13 +100,16 @@ class _TestProcess(ServerTestCase):
         data = str(id(self))
 
         with (yield from self.connect()) as conn:
-            process = yield from conn.create_process()
+            process = yield from conn.create_process(env={'TEST': 'test'})
 
             process.stdin.write(data)
             process.stdin.write_eof()
 
             result = yield from process.wait()
 
+        self.assertEqual(result.env, {'TEST': 'test'})
+        self.assertEqual(result.command, None)
+        self.assertEqual(result.subsystem, None)
         self.assertEqual(result.exit_status, None)
         self.assertEqual(result.exit_signal, None)
         self.assertEqual(result.stdout, data)
@@ -102,6 +129,27 @@ class _TestProcess(ServerTestCase):
 
             result = yield from process.wait()
 
+        self.assertEqual(result.command, 'echo')
+        self.assertEqual(result.subsystem, None)
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_subsystem(self):
+        """Test starting a remote subsystem"""
+
+        data = str(id(self))
+
+        with (yield from self.connect()) as conn:
+            process = yield from conn.create_process(subsystem='echo')
+
+            process.stdin.write(data)
+            process.stdin.write_eof()
+
+            result = yield from process.wait()
+
+        self.assertEqual(result.command, None)
+        self.assertEqual(result.subsystem, 'echo')
         self.assertEqual(result.stdout, data)
         self.assertEqual(result.stderr, data)
 
@@ -132,12 +180,38 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(stdout_data, data)
         self.assertEqual(stderr_data, data)
 
+    @asynctest
+    def test_env(self):
+        """Test sending environment"""
+
+        with (yield from self.connect()) as conn:
+            process = yield from conn.create_process('env',
+                                                     env={'TEST': 'test'})
+            result = yield from process.wait()
+
+        self.assertEqual(result.stdout, 'test')
+
+    @asynctest
+    def test_terminal_info(self):
+        """Test sending terminal information"""
+
+        modes = {asyncssh.PTY_OP_OSPEED: 9600}
+
+        with (yield from self.connect()) as conn:
+            process = yield from conn.create_process('term', term_type='ansi',
+                                                     term_size=(80, 24),
+                                                     term_modes=modes)
+            result = yield from process.wait()
+
+        self.assertEqual(result.stdout, "('ansi', (80, 24, 0, 0), 9600)")
+
     @asynctest
     def test_change_terminal_size(self):
         """Test changing terminal size"""
 
         with (yield from self.connect()) as conn:
-            process = yield from conn.create_process('term', term_type='ansi')
+            process = yield from conn.create_process('term_size',
+                                                     term_type='ansi')
             process.change_terminal_size(80, 24)
             result = yield from process.wait()
 
@@ -195,6 +269,8 @@ class _TestProcess(ServerTestCase):
             result = yield from conn.run('exit_status')
 
         self.assertEqual(result.exit_status, 1)
+        self.assertEqual(result.stdout, '')
+        self.assertEqual(result.stderr, 'Exiting with status 1')
 
     @asynctest
     def test_raise_on_exit_status(self):
@@ -202,8 +278,12 @@ class _TestProcess(ServerTestCase):
 
         with (yield from self.connect()) as conn:
             with self.assertRaises(asyncssh.ProcessError) as exc:
-                yield from conn.run('exit_status', check=True)
+                yield from conn.run('exit_status', env={'TEST': 'test'},
+                                    check=True)
 
+        self.assertEqual(exc.exception.env, {'TEST': 'test'})
+        self.assertEqual(exc.exception.command, 'exit_status')
+        self.assertEqual(exc.exception.subsystem, None)
         self.assertEqual(exc.exception.exit_status, 1)
         self.assertEqual(exc.exception.reason,
                          'Process exited with non-zero exit status 1')
@@ -236,6 +316,76 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(exc.exception.reason,
                          'Process exited with signal HUP')
 
+    @asynctest
+    def test_split_unicode(self):
+        """Test Unicode split across blocks"""
+
+        data = '\u2000test\u2000'
+
+        with open('stdin', 'w', encoding='utf-8') as file:
+            file.write(data)
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', stdin='stdin', bufsize=2)
+
+        self.assertEqual(result.stdout, data)
+
+    @asynctest
+    def test_invalid_unicode(self):
+        """Test invalid Unicode data"""
+
+        data = b'\xfftest'
+
+        with open('stdin', 'wb') as file:
+            file.write(data)
+
+        with (yield from self.connect()) as conn:
+            with self.assertRaises(asyncssh.DisconnectError):
+                yield from conn.run('echo', stdin='stdin')
+
+    @asynctest
+    def test_incomplete_unicode(self):
+        """Test incomplete Unicode data"""
+
+        data = '\u2000'.encode('utf-8')[:2]
+
+        with open('stdin', 'wb') as file:
+            file.write(data)
+
+        with (yield from self.connect()) as conn:
+            with self.assertRaises(asyncssh.DisconnectError):
+                yield from conn.run('echo', stdin='stdin')
+
+    @asynctest
+    def test_disconnect(self):
+        """Test collecting output from a disconnected channel"""
+
+        data = str(id(self))
+
+        with (yield from self.connect()) as conn:
+            process = yield from conn.create_process()
+
+            process.stdin.write(data)
+            process.send_signal('ABRT')
+
+            result = yield from process.wait()
+
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_unknown_action(self):
+        """Test unknown action"""
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('unknown')
+
+        self.assertEqual(result.exit_status, 255)
+
+
+class _TestProcessRedirection(_TestProcess):
+    """Unit tests for AsyncSSH process I/O redirection"""
+
     @asynctest
     def test_input(self):
         """Test with input from a string"""
@@ -325,53 +475,35 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(result.stderr, data)
 
     @asynctest
-    def test_stdin_pipe(self):
-        """Test with stdin redirected to a pipe"""
+    def test_stdin_stringio(self):
+        """Test with stdin redirected to a StringIO object"""
 
         data = str(id(self))
 
-        rpipe, wpipe = os.pipe()
+        with open('stdin', 'w') as file:
+            file.write(data)
 
-        os.write(wpipe, data.encode())
-        os.close(wpipe)
+        file = io.StringIO(data)
 
         with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', stdin=rpipe)
+            result = yield from conn.run('echo', stdin=file)
 
         self.assertEqual(result.stdout, data)
         self.assertEqual(result.stderr, data)
 
     @asynctest
-    def test_stdin_binary_pipe(self):
-        """Test with stdin redirected to a pipe in binary mode"""
-
-        data = str(id(self)).encode() + b'\xff'
-
-        rpipe, wpipe = os.pipe()
-
-        os.write(wpipe, data)
-        os.close(wpipe)
-
-        with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', stdin=rpipe,
-                                         encoding=None)
-
-        self.assertEqual(result.stdout, data)
-        self.assertEqual(result.stderr, data)
-
-    @asynctest
-    def test_stdin_socketpair(self):
-        """Test with stdin redirected to a socketpair"""
+    def test_stdin_bytesio(self):
+        """Test with stdin redirected to a BytesIO object"""
 
         data = str(id(self))
 
-        sock1, sock2 = socket.socketpair()
+        with open('stdin', 'w') as file:
+            file.write(data)
 
-        sock1.send(data.encode())
-        sock1.close()
+        file = io.BytesIO(data.encode('ascii'))
 
         with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', stdin=sock2)
+            result = yield from conn.run('echo', stdin=file)
 
         self.assertEqual(result.stdout, data)
         self.assertEqual(result.stderr, data)
@@ -390,30 +522,6 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(result.stdout, data)
         self.assertEqual(result.stderr, data)
 
-    @asynctest
-    def test_change_stdin(self):
-        """Test changing stdin of an open process"""
-
-        sock1, sock2 = socket.socketpair()
-        sock3, sock4 = socket.socketpair()
-
-        sock1.send(b'xxx')
-        sock3.send(b'yyy')
-
-        with (yield from self.connect()) as conn:
-            process = yield from conn.create_process(stdin=sock2)
-
-            yield from asyncio.sleep(0.1)
-            yield from process.redirect_stdin(sock4)
-
-            sock1.close()
-            sock3.close()
-
-            result = yield from process.wait()
-
-        self.assertEqual(result.stdout, 'xxxyyy')
-        self.assertEqual(result.stderr, 'xxxyyy')
-
     @asynctest
     def test_stdout_devnull(self):
         """Test with stdout redirected to DEVNULL"""
@@ -498,57 +606,57 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(result.stderr, data)
 
     @asynctest
-    def test_stdout_pipe(self):
-        """Test with stdout redirected to a pipe"""
+    def test_stdout_stringio(self):
+        """Test with stdout redirected to a StringIO"""
+
+        class _StringIOTest(io.StringIO):
+            """Test class for StringIO which preserves output after close"""
+
+            def __init__(self):
+                super().__init__()
+                self.output = None
+
+            def close(self):
+                if self.output is None:
+                    self.output = self.getvalue()
+                    super().close()
 
         data = str(id(self))
 
-        rpipe, wpipe = os.pipe()
+        file = _StringIOTest()
 
         with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', input=data, stdout=wpipe)
+            result = yield from conn.run('echo', input=data, stdout=file)
 
-        stdout_data = os.read(rpipe, 1024)
-        os.close(rpipe)
-
-        self.assertEqual(stdout_data.decode(), data)
+        self.assertEqual(file.output, data)
         self.assertEqual(result.stdout, '')
         self.assertEqual(result.stderr, data)
 
     @asynctest
-    def test_stdout_binary_pipe(self):
-        """Test with stdout redirected to a pipe in binary mode"""
+    def test_stdout_bytesio(self):
+        """Test with stdout redirected to a BytesIO"""
 
-        data = str(id(self)).encode() + b'\xff'
+        class _BytesIOTest(io.BytesIO):
+            """Test class for BytesIO which preserves output after close"""
 
-        rpipe, wpipe = os.pipe()
+            def __init__(self):
+                super().__init__()
+                self.output = None
 
-        with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', input=data, stdout=wpipe,
-                                         encoding=None)
-
-        stdout_data = os.read(rpipe, 1024)
-        os.close(rpipe)
-
-        self.assertEqual(stdout_data, data)
-        self.assertEqual(result.stdout, b'')
-        self.assertEqual(result.stderr, data)
-
-    @asynctest
-    def test_stdout_socketpair(self):
-        """Test with stdout redirected to a socketpair"""
+            def close(self):
+                if self.output is None:
+                    self.output = self.getvalue()
+                    super().close()
 
         data = str(id(self))
 
-        sock1, sock2 = socket.socketpair()
+        file = _BytesIOTest()
 
         with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', input=data, stdout=sock1)
+            result = yield from conn.run('echo', input=data, stdout=file)
 
-        stdout_data = sock2.recv(1024)
-        sock2.close()
-
-        self.assertEqual(stdout_data.decode(), data)
+        self.assertEqual(file.output, data.encode('ascii'))
+        self.assertEqual(result.stdout, '')
         self.assertEqual(result.stderr, data)
 
     @asynctest
@@ -593,6 +701,31 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(result.stdout, 'yyy')
         self.assertEqual(result.stderr, 'xxxyyy')
 
+    @asynctest
+    def test_change_stdin_process(self):
+        """Test changing stdin of an open process reading from another"""
+
+        data = str(id(self))
+
+        with (yield from self.connect()) as conn:
+            with (yield from conn.create_process()) as proc2:
+                proc1 = yield from conn.create_process(stdout=proc2.stdin)
+
+                proc1.stdin.write(data)
+                yield from asyncio.sleep(0.1)
+
+                yield from proc2.redirect_stdin(asyncssh.PIPE)
+                proc2.stdin.write(data)
+                yield from asyncio.sleep(0.1)
+
+                yield from proc2.redirect_stdin(proc1.stdout)
+                proc1.stdin.write_eof()
+
+                result = yield from proc2.wait()
+
+        self.assertEqual(result.stdout, data+data)
+        self.assertEqual(result.stderr, data+data)
+
     @asynctest
     def test_change_stdout_process(self):
         """Test changing stdout of an open process sending to another"""
@@ -630,6 +763,42 @@ class _TestProcess(ServerTestCase):
 
         self.assertEqual(result.stdout, data+data)
 
+    @asynctest
+    def test_server_redirect_stdin(self):
+        """Test redirect on server of stdin"""
+
+        data = str(id(self))
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('redirect_stdin', input=data)
+
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, '')
+
+    @asynctest
+    def test_server_redirect_stdout(self):
+        """Test redirect on server of stdout"""
+
+        data = str(id(self))
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('redirect_stdout', input=data)
+
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, '')
+
+    @asynctest
+    def test_server_redirect_stderr(self):
+        """Test redirect on server of stderr"""
+
+        data = str(id(self))
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('redirect_stderr', input=data)
+
+        self.assertEqual(result.stdout, '')
+        self.assertEqual(result.stderr, data)
+
     @asynctest
     def test_pause_file_reader(self):
         """Test pausing and resuming reading from a file"""
@@ -645,55 +814,6 @@ class _TestProcess(ServerTestCase):
 
         self.assertEqual(result.stdout, data)
 
-    @asynctest
-    def test_pause_pipe_reader(self):
-        """Test pausing and resuming reading from a socketpair"""
-
-        data = 4*1024*1024*'*'
-
-        sock1, sock2 = socket.socketpair()
-
-        _, writer = yield from asyncio.open_unix_connection(sock=sock1)
-        writer.write(data.encode())
-        writer.close()
-
-        with (yield from self.connect()) as conn:
-            result = yield from conn.run('delay', stdin=sock2,
-                                         stderr=asyncssh.DEVNULL)
-
-        self.assertEqual(result.stdout, data)
-
-    @asynctest
-    def test_pause_pipe_writer(self):
-        """Test pausing and resuming writing to a socketpair"""
-
-        data = 4*1024*1024*'*'
-
-        rsock1, wsock1 = socket.socketpair()
-        rsock2, wsock2 = socket.socketpair()
-
-        reader1, writer1 = yield from asyncio.open_unix_connection(sock=rsock1)
-        reader2, writer2 = yield from asyncio.open_unix_connection(sock=rsock2)
-
-        with (yield from self.connect()) as conn:
-            process = yield from conn.create_process(input=data)
-
-            yield from asyncio.sleep(1)
-
-            yield from process.redirect_stdout(wsock1)
-            yield from process.redirect_stderr(wsock2)
-
-            stdout_data, stderr_data = \
-                yield from asyncio.gather(reader1.read(), reader2.read())
-
-            writer1.close()
-            writer2.close()
-
-            yield from process.wait()
-
-        self.assertEqual(stdout_data.decode(), data)
-        self.assertEqual(stderr_data.decode(), data)
-
     @asynctest
     def test_pause_process_reader(self):
         """Test pausing and resuming reading from another SSH process"""
@@ -776,68 +896,232 @@ class _TestProcess(ServerTestCase):
         self.assertEqual(result.stdout, data+data)
         self.assertEqual(result.stderr, data+data)
 
-    @asynctest
-    def test_split_unicode(self):
-        """Test Unicode split across blocks"""
 
-        data = '\u2000test\u2000'
-
-        with open('stdin', 'w', encoding='utf-8') as file:
-            file.write(data)
-
-        with (yield from self.connect()) as conn:
-            result = yield from conn.run('echo', stdin='stdin', bufsize=2)
-
-        self.assertEqual(result.stdout, data)
+@unittest.skipIf(sys.platform == 'win32', 'skip pipe tests on Windows')
+class _TestProcessPipes(_TestProcess):
+    """Unit tests for AsyncSSH process I/O using pipes"""
 
     @asynctest
-    def test_invalid_unicode(self):
-        """Test invalid Unicode data"""
-
-        data = b'\xfftest'
-
-        with open('stdin', 'wb') as file:
-            file.write(data)
-
-        with (yield from self.connect()) as conn:
-            with self.assertRaises(asyncssh.DisconnectError):
-                yield from conn.run('echo', stdin='stdin')
-
-    @asynctest
-    def test_incomplete_unicode(self):
-        """Test incomplete Unicode data"""
-
-        data = '\u2000'.encode('utf-8')[:2]
-
-        with open('stdin', 'wb') as file:
-            file.write(data)
-
-        with (yield from self.connect()) as conn:
-            with self.assertRaises(asyncssh.DisconnectError):
-                yield from conn.run('echo', stdin='stdin')
-
-    @asynctest
-    def test_disconnect(self):
-        """Test collecting output from a disconnected channel"""
+    def test_stdin_pipe(self):
+        """Test with stdin redirected to a pipe"""
 
         data = str(id(self))
 
+        rpipe, wpipe = os.pipe()
+
+        os.write(wpipe, data.encode())
+        os.close(wpipe)
+
         with (yield from self.connect()) as conn:
-            process = yield from conn.create_process()
-
-            process.stdin.write(data)
-            process.send_signal('ABRT')
-
-            result = yield from process.wait()
+            result = yield from conn.run('echo', stdin=rpipe)
 
         self.assertEqual(result.stdout, data)
         self.assertEqual(result.stderr, data)
 
     @asynctest
-    def test_unknown_action(self):
-        """Test unknown action"""
+    def test_stdin_text_pipe(self):
+        """Test with stdin redirected to a pipe in text mode"""
+
+        data = str(id(self))
+
+        rpipe, wpipe = os.pipe()
+
+        rpipe = os.fdopen(rpipe, 'r')
+        wpipe = os.fdopen(wpipe, 'w')
+
+        wpipe.write(data)
+        wpipe.close()
 
         with (yield from self.connect()) as conn:
-            result = yield from conn.run('unknown')
+            result = yield from conn.run('echo', stdin=rpipe)
 
-        self.assertEqual(result.exit_status, 255)
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_stdin_binary_pipe(self):
+        """Test with stdin redirected to a pipe in binary mode"""
+
+        data = str(id(self)).encode() + b'\xff'
+
+        rpipe, wpipe = os.pipe()
+
+        os.write(wpipe, data)
+        os.close(wpipe)
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', stdin=rpipe,
+                                         encoding=None)
+
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_stdout_pipe(self):
+        """Test with stdout redirected to a pipe"""
+
+        data = str(id(self))
+
+        rpipe, wpipe = os.pipe()
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', input=data, stdout=wpipe)
+
+        stdout_data = os.read(rpipe, 1024)
+        os.close(rpipe)
+
+        self.assertEqual(stdout_data.decode(), data)
+        self.assertEqual(result.stdout, '')
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_stdout_text_pipe(self):
+        """Test with stdout redirected to a pipe in text mode"""
+
+        data = str(id(self))
+
+        rpipe, wpipe = os.pipe()
+
+        rpipe = os.fdopen(rpipe, 'r')
+        wpipe = os.fdopen(wpipe, 'w')
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', input=data, stdout=wpipe)
+
+        stdout_data = rpipe.read(1024)
+        rpipe.close()
+
+        self.assertEqual(stdout_data, data)
+        self.assertEqual(result.stdout, '')
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_stdout_binary_pipe(self):
+        """Test with stdout redirected to a pipe in binary mode"""
+
+        data = str(id(self)).encode() + b'\xff'
+
+        rpipe, wpipe = os.pipe()
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', input=data, stdout=wpipe,
+                                         encoding=None)
+
+        stdout_data = os.read(rpipe, 1024)
+        os.close(rpipe)
+
+        self.assertEqual(stdout_data, data)
+        self.assertEqual(result.stdout, b'')
+        self.assertEqual(result.stderr, data)
+
+
+@unittest.skipIf(sys.platform == 'win32', 'skip socketpair tests on Windows')
+class _TestProcessSocketPair(_TestProcess):
+    """Unit tests for AsyncSSH process I/O using socketpair"""
+
+    @asynctest
+    def test_stdin_socketpair(self):
+        """Test with stdin redirected to a socketpair"""
+
+        data = str(id(self))
+
+        sock1, sock2 = socket.socketpair()
+
+        sock1.send(data.encode())
+        sock1.close()
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', stdin=sock2)
+
+        self.assertEqual(result.stdout, data)
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_change_stdin(self):
+        """Test changing stdin of an open process"""
+
+        sock1, sock2 = socket.socketpair()
+        sock3, sock4 = socket.socketpair()
+
+        sock1.send(b'xxx')
+        sock3.send(b'yyy')
+
+        with (yield from self.connect()) as conn:
+            process = yield from conn.create_process(stdin=sock2)
+
+            yield from asyncio.sleep(0.1)
+            yield from process.redirect_stdin(sock4)
+
+            sock1.close()
+            sock3.close()
+
+            result = yield from process.wait()
+
+        self.assertEqual(result.stdout, 'xxxyyy')
+        self.assertEqual(result.stderr, 'xxxyyy')
+
+    @asynctest
+    def test_stdout_socketpair(self):
+        """Test with stdout redirected to a socketpair"""
+
+        data = str(id(self))
+
+        sock1, sock2 = socket.socketpair()
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('echo', input=data, stdout=sock1)
+
+        stdout_data = sock2.recv(1024)
+        sock2.close()
+
+        self.assertEqual(stdout_data.decode(), data)
+        self.assertEqual(result.stderr, data)
+
+    @asynctest
+    def test_pause_socketpair_reader(self):
+        """Test pausing and resuming reading from a socketpair"""
+
+        data = 4*1024*1024*'*'
+
+        sock1, sock2 = socket.socketpair()
+
+        _, writer = yield from asyncio.open_unix_connection(sock=sock1)
+        writer.write(data.encode())
+        writer.close()
+
+        with (yield from self.connect()) as conn:
+            result = yield from conn.run('delay', stdin=sock2,
+                                         stderr=asyncssh.DEVNULL)
+
+        self.assertEqual(result.stdout, data)
+
+    @asynctest
+    def test_pause_socketpair_writer(self):
+        """Test pausing and resuming writing to a socketpair"""
+
+        data = 4*1024*1024*'*'
+
+        rsock1, wsock1 = socket.socketpair()
+        rsock2, wsock2 = socket.socketpair()
+
+        reader1, writer1 = yield from asyncio.open_unix_connection(sock=rsock1)
+        reader2, writer2 = yield from asyncio.open_unix_connection(sock=rsock2)
+
+        with (yield from self.connect()) as conn:
+            process = yield from conn.create_process(input=data)
+
+            yield from asyncio.sleep(1)
+
+            yield from process.redirect_stdout(wsock1)
+            yield from process.redirect_stderr(wsock2)
+
+            stdout_data, stderr_data = \
+                yield from asyncio.gather(reader1.read(), reader2.read())
+
+            writer1.close()
+            writer2.close()
+
+            yield from process.wait()
+
+        self.assertEqual(stdout_data.decode(), data)
+        self.assertEqual(stderr_data.decode(), data)
